/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.server.metrics.prometheus;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.errorprone.annotations.CanIgnoreReturnValue;
import com.google.errorprone.annotations.Var;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import javax.annotation.CheckReturnValue;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import javax.annotation.concurrent.Immutable;
import javax.annotation.concurrent.NotThreadSafe;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;
import org.immutables.value.Generated;

/**
 * Immutable implementation of {@link QueryResult.Data}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new QueryResult.Data.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableData.of()}.
 */
@Generated(from = "QueryResult.Data", generator = "Immutables")
@SuppressWarnings({"all"})
@SuppressFBWarnings
@ParametersAreNonnullByDefault
@javax.annotation.Generated("org.immutables.processor.ProxyProcessor")
@Immutable
@CheckReturnValue
final class ImmutableData implements QueryResult.Data {
  private final String resultType;
  private final List<QueryResult.Data.Result> result;

  private ImmutableData(
      String resultType,
      Iterable<? extends QueryResult.Data.Result> result) {
    this.resultType = resultType;
    this.result = createUnmodifiableList(false, createSafeList(result, true, false));
  }

  private ImmutableData(
      ImmutableData original,
      String resultType,
      List<QueryResult.Data.Result> result) {
    this.resultType = resultType;
    this.result = result;
  }

  /**
   * @return The value of the {@code resultType} attribute
   */
  @JsonProperty("resultType")
  @Override
  public String getResultType() {
    return resultType;
  }

  /**
   * @return The value of the {@code result} attribute
   */
  @JsonProperty("result")
  @Override
  public List<QueryResult.Data.Result> getResult() {
    return result;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link QueryResult.Data#getResultType() resultType} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for resultType (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableData withResultType(String value) {
    if (Objects.equals(this.resultType, value)) return this;
    return validate(new ImmutableData(this, value, this.result));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link QueryResult.Data#getResult() result}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableData withResult(QueryResult.Data.Result... elements) {
    List<QueryResult.Data.Result> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableData(this, this.resultType, newValue));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link QueryResult.Data#getResult() result}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of result elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableData withResult(Iterable<? extends QueryResult.Data.Result> elements) {
    if (this.result == elements) return this;
    List<QueryResult.Data.Result> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableData(this, this.resultType, newValue));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableData} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(@Nullable Object another) {
    if (this == another) return true;
    return another instanceof ImmutableData
        && equalTo((ImmutableData) another);
  }

  private boolean equalTo(ImmutableData another) {
    return Objects.equals(resultType, another.resultType)
        && result.equals(another.result);
  }

  /**
   * Computes a hash code from attributes: {@code resultType}, {@code result}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    @Var int h = 5381;
    h += (h << 5) + Objects.hashCode(resultType);
    h += (h << 5) + result.hashCode();
    return h;
  }


  /**
   * Prints the immutable value {@code Data} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "Data{"
        + "resultType=" + resultType
        + ", result=" + result
        + "}";
  }

  /**
   * Construct a new immutable {@code Data} instance.
   * @param resultType The value for the {@code resultType} attribute
   * @param result The value for the {@code result} attribute
   * @return An immutable Data instance
   */
  public static QueryResult.Data of(String resultType, List<QueryResult.Data.Result> result) {
    return of(resultType, (Iterable<? extends QueryResult.Data.Result>) result);
  }

  /**
   * Construct a new immutable {@code Data} instance.
   * @param resultType The value for the {@code resultType} attribute
   * @param result The value for the {@code result} attribute
   * @return An immutable Data instance
   */
  public static QueryResult.Data of(String resultType, Iterable<? extends QueryResult.Data.Result> result) {
    return validate(new ImmutableData(resultType, result));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableData validate(ImmutableData instance) {
    Set<ConstraintViolation<ImmutableData>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link QueryResult.Data} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Data instance
   */
  public static QueryResult.Data copyOf(QueryResult.Data instance) {
    if (instance instanceof ImmutableData) {
      return (ImmutableData) instance;
    }
    return new QueryResult.Data.Builder()
        .createFrom(instance)
        .build();
  }

  /**
   * Builds instances of type {@link QueryResult.Data Data}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  @Generated(from = "QueryResult.Data", generator = "Immutables")
  @NotThreadSafe
  public static class Builder {
    private @Nullable String resultType;
    private List<QueryResult.Data.Result> result = new ArrayList<QueryResult.Data.Result>();

    /**
     * Creates a builder for {@link QueryResult.Data Data} instances.
     * <pre>
     * new QueryResult.Data.Builder()
     *    .resultType(String | null) // nullable {@link QueryResult.Data#getResultType() resultType}
     *    .addResult|addAllResult(io.syndesis.server.metrics.prometheus.QueryResult.Data.Result) // {@link QueryResult.Data#getResult() result} elements
     *    .build();
     * </pre>
     */
    public Builder() {
      if (!(this instanceof QueryResult.Data.Builder)) {
        throw new UnsupportedOperationException("Use: new QueryResult.Data.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code Data} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Data.Builder createFrom(QueryResult.Data instance) {
      Objects.requireNonNull(instance, "instance");
      String resultTypeValue = instance.getResultType();
      if (resultTypeValue != null) {
        resultType(resultTypeValue);
      }
      addAllResult(instance.getResult());
      return (QueryResult.Data.Builder) this;
    }

    /**
     * Initializes the value for the {@link QueryResult.Data#getResultType() resultType} attribute.
     * @param resultType The value for resultType (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    @JsonProperty("resultType")
    public final QueryResult.Data.Builder resultType(String resultType) {
      this.resultType = resultType;
      return (QueryResult.Data.Builder) this;
    }

    /**
     * Adds one element to {@link QueryResult.Data#getResult() result} list.
     * @param element A result element
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Data.Builder addResult(QueryResult.Data.Result element) {
      this.result.add(Objects.requireNonNull(element, "result element"));
      return (QueryResult.Data.Builder) this;
    }

    /**
     * Adds elements to {@link QueryResult.Data#getResult() result} list.
     * @param elements An array of result elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Data.Builder addResult(QueryResult.Data.Result... elements) {
      for (QueryResult.Data.Result element : elements) {
        this.result.add(Objects.requireNonNull(element, "result element"));
      }
      return (QueryResult.Data.Builder) this;
    }


    /**
     * Sets or replaces all elements for {@link QueryResult.Data#getResult() result} list.
     * @param elements An iterable of result elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    @JsonProperty("result")
    public final QueryResult.Data.Builder result(Iterable<? extends QueryResult.Data.Result> elements) {
      this.result.clear();
      return addAllResult(elements);
    }

    /**
     * Adds elements to {@link QueryResult.Data#getResult() result} list.
     * @param elements An iterable of result elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Data.Builder addAllResult(Iterable<? extends QueryResult.Data.Result> elements) {
      for (QueryResult.Data.Result element : elements) {
        this.result.add(Objects.requireNonNull(element, "result element"));
      }
      return (QueryResult.Data.Builder) this;
    }

    /**
     * Builds a new {@link QueryResult.Data Data}.
     * @return An immutable instance of Data
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public QueryResult.Data build() {
      return ImmutableData.validate(new ImmutableData(null, resultType, createUnmodifiableList(true, result)));
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }
}
