/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.server.metrics.prometheus;

import com.google.errorprone.annotations.CanIgnoreReturnValue;
import com.google.errorprone.annotations.Var;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import javax.annotation.CheckReturnValue;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import javax.annotation.concurrent.Immutable;
import javax.annotation.concurrent.NotThreadSafe;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;
import org.immutables.value.Generated;

/**
 * Immutable implementation of {@link HttpQuery}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new HttpQuery.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableHttpQuery.of()}.
 */
@Generated(from = "HttpQuery", generator = "Immutables")
@SuppressWarnings({"all"})
@SuppressFBWarnings
@ParametersAreNonnullByDefault
@javax.annotation.Generated("org.immutables.processor.ProxyProcessor")
@Immutable
@CheckReturnValue
final class ImmutableHttpQuery implements HttpQuery {
  private final String host;
  private final @Nullable String function;
  private final @Nullable String aggregationOperator;
  private final List<String> aggregationOperatorParameters;
  private final String metric;
  private final List<HttpQuery.LabelValue> labelValues;
  private final @Nullable String range;
  private final List<String> withoutLabels;
  private final List<String> byLabels;

  private ImmutableHttpQuery(
      String host,
      Optional<String> function,
      Optional<String> aggregationOperator,
      Iterable<String> aggregationOperatorParameters,
      String metric,
      Iterable<? extends HttpQuery.LabelValue> labelValues,
      Optional<String> range,
      Iterable<String> withoutLabels,
      Iterable<String> byLabels) {
    this.host = host;
    this.function = function.orElse(null);
    this.aggregationOperator = aggregationOperator.orElse(null);
    this.aggregationOperatorParameters = createUnmodifiableList(false, createSafeList(aggregationOperatorParameters, true, false));
    this.metric = metric;
    this.labelValues = createUnmodifiableList(false, createSafeList(labelValues, true, false));
    this.range = range.orElse(null);
    this.withoutLabels = createUnmodifiableList(false, createSafeList(withoutLabels, true, false));
    this.byLabels = createUnmodifiableList(false, createSafeList(byLabels, true, false));
  }

  private ImmutableHttpQuery(
      ImmutableHttpQuery original,
      String host,
      @Nullable String function,
      @Nullable String aggregationOperator,
      List<String> aggregationOperatorParameters,
      String metric,
      List<HttpQuery.LabelValue> labelValues,
      @Nullable String range,
      List<String> withoutLabels,
      List<String> byLabels) {
    this.host = host;
    this.function = function;
    this.aggregationOperator = aggregationOperator;
    this.aggregationOperatorParameters = aggregationOperatorParameters;
    this.metric = metric;
    this.labelValues = labelValues;
    this.range = range;
    this.withoutLabels = withoutLabels;
    this.byLabels = byLabels;
  }

  /**
   * @return The value of the {@code host} attribute
   */
  @Override
  public String getHost() {
    return host;
  }

  /**
   * @return The value of the {@code function} attribute
   */
  @Override
  public Optional<String> getFunction() {
    return Optional.ofNullable(function);
  }

  /**
   * @return The value of the {@code aggregationOperator} attribute
   */
  @Override
  public Optional<String> getAggregationOperator() {
    return Optional.ofNullable(aggregationOperator);
  }

  /**
   * @return The value of the {@code aggregationOperatorParameters} attribute
   */
  @Override
  public List<String> getAggregationOperatorParameters() {
    return aggregationOperatorParameters;
  }

  /**
   * @return The value of the {@code metric} attribute
   */
  @Override
  public String getMetric() {
    return metric;
  }

  /**
   * @return The value of the {@code labelValues} attribute
   */
  @Override
  public List<HttpQuery.LabelValue> getLabelValues() {
    return labelValues;
  }

  /**
   * @return The value of the {@code range} attribute
   */
  @Override
  public Optional<String> getRange() {
    return Optional.ofNullable(range);
  }

  /**
   * @return The value of the {@code withoutLabels} attribute
   */
  @Override
  public List<String> getWithoutLabels() {
    return withoutLabels;
  }

  /**
   * @return The value of the {@code byLabels} attribute
   */
  @Override
  public List<String> getByLabels() {
    return byLabels;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link HttpQuery#getHost() host} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for host (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableHttpQuery withHost(String value) {
    if (Objects.equals(this.host, value)) return this;
    return validate(new ImmutableHttpQuery(
        this,
        value,
        this.function,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        this.metric,
        this.labelValues,
        this.range,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link HttpQuery#getFunction() function} attribute.
   * @param value The value for function
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withFunction(String value) {
    @Nullable String newValue = Objects.requireNonNull(value, "function");
    if (Objects.equals(this.function, newValue)) return this;
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        newValue,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        this.metric,
        this.labelValues,
        this.range,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link HttpQuery#getFunction() function} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for function
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withFunction(Optional<String> optional) {
    @Nullable String value = optional.orElse(null);
    if (Objects.equals(this.function, value)) return this;
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        value,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        this.metric,
        this.labelValues,
        this.range,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link HttpQuery#getAggregationOperator() aggregationOperator} attribute.
   * @param value The value for aggregationOperator
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withAggregationOperator(String value) {
    @Nullable String newValue = Objects.requireNonNull(value, "aggregationOperator");
    if (Objects.equals(this.aggregationOperator, newValue)) return this;
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        newValue,
        this.aggregationOperatorParameters,
        this.metric,
        this.labelValues,
        this.range,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link HttpQuery#getAggregationOperator() aggregationOperator} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for aggregationOperator
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withAggregationOperator(Optional<String> optional) {
    @Nullable String value = optional.orElse(null);
    if (Objects.equals(this.aggregationOperator, value)) return this;
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        value,
        this.aggregationOperatorParameters,
        this.metric,
        this.labelValues,
        this.range,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link HttpQuery#getAggregationOperatorParameters() aggregationOperatorParameters}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withAggregationOperatorParameters(String... elements) {
    List<String> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        this.aggregationOperator,
        newValue,
        this.metric,
        this.labelValues,
        this.range,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link HttpQuery#getAggregationOperatorParameters() aggregationOperatorParameters}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of aggregationOperatorParameters elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withAggregationOperatorParameters(Iterable<String> elements) {
    if (this.aggregationOperatorParameters == elements) return this;
    List<String> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        this.aggregationOperator,
        newValue,
        this.metric,
        this.labelValues,
        this.range,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link HttpQuery#getMetric() metric} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for metric (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableHttpQuery withMetric(String value) {
    if (Objects.equals(this.metric, value)) return this;
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        value,
        this.labelValues,
        this.range,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link HttpQuery#getLabelValues() labelValues}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withLabelValues(HttpQuery.LabelValue... elements) {
    List<HttpQuery.LabelValue> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        this.metric,
        newValue,
        this.range,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link HttpQuery#getLabelValues() labelValues}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of labelValues elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withLabelValues(Iterable<? extends HttpQuery.LabelValue> elements) {
    if (this.labelValues == elements) return this;
    List<HttpQuery.LabelValue> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        this.metric,
        newValue,
        this.range,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link HttpQuery#getRange() range} attribute.
   * @param value The value for range
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withRange(String value) {
    @Nullable String newValue = Objects.requireNonNull(value, "range");
    if (Objects.equals(this.range, newValue)) return this;
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        this.metric,
        this.labelValues,
        newValue,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link HttpQuery#getRange() range} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for range
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withRange(Optional<String> optional) {
    @Nullable String value = optional.orElse(null);
    if (Objects.equals(this.range, value)) return this;
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        this.metric,
        this.labelValues,
        value,
        this.withoutLabels,
        this.byLabels));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link HttpQuery#getWithoutLabels() withoutLabels}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withWithoutLabels(String... elements) {
    List<String> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        this.metric,
        this.labelValues,
        this.range,
        newValue,
        this.byLabels));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link HttpQuery#getWithoutLabels() withoutLabels}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of withoutLabels elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withWithoutLabels(Iterable<String> elements) {
    if (this.withoutLabels == elements) return this;
    List<String> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        this.metric,
        this.labelValues,
        this.range,
        newValue,
        this.byLabels));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link HttpQuery#getByLabels() byLabels}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withByLabels(String... elements) {
    List<String> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        this.metric,
        this.labelValues,
        this.range,
        this.withoutLabels,
        newValue));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link HttpQuery#getByLabels() byLabels}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of byLabels elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableHttpQuery withByLabels(Iterable<String> elements) {
    if (this.byLabels == elements) return this;
    List<String> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableHttpQuery(
        this,
        this.host,
        this.function,
        this.aggregationOperator,
        this.aggregationOperatorParameters,
        this.metric,
        this.labelValues,
        this.range,
        this.withoutLabels,
        newValue));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableHttpQuery} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(@Nullable Object another) {
    if (this == another) return true;
    return another instanceof ImmutableHttpQuery
        && equalTo((ImmutableHttpQuery) another);
  }

  private boolean equalTo(ImmutableHttpQuery another) {
    return Objects.equals(host, another.host)
        && Objects.equals(function, another.function)
        && Objects.equals(aggregationOperator, another.aggregationOperator)
        && aggregationOperatorParameters.equals(another.aggregationOperatorParameters)
        && Objects.equals(metric, another.metric)
        && labelValues.equals(another.labelValues)
        && Objects.equals(range, another.range)
        && withoutLabels.equals(another.withoutLabels)
        && byLabels.equals(another.byLabels);
  }

  /**
   * Computes a hash code from attributes: {@code host}, {@code function}, {@code aggregationOperator}, {@code aggregationOperatorParameters}, {@code metric}, {@code labelValues}, {@code range}, {@code withoutLabels}, {@code byLabels}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    @Var int h = 5381;
    h += (h << 5) + Objects.hashCode(host);
    h += (h << 5) + Objects.hashCode(function);
    h += (h << 5) + Objects.hashCode(aggregationOperator);
    h += (h << 5) + aggregationOperatorParameters.hashCode();
    h += (h << 5) + Objects.hashCode(metric);
    h += (h << 5) + labelValues.hashCode();
    h += (h << 5) + Objects.hashCode(range);
    h += (h << 5) + withoutLabels.hashCode();
    h += (h << 5) + byLabels.hashCode();
    return h;
  }


  /**
   * Prints the immutable value {@code HttpQuery} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder("HttpQuery{");
    if (host != null) {
      builder.append("host=").append(host);
    }
    if (function != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("function=").append(function);
    }
    if (aggregationOperator != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("aggregationOperator=").append(aggregationOperator);
    }
    if (builder.length() > 10) builder.append(", ");
    builder.append("aggregationOperatorParameters=").append(aggregationOperatorParameters);
    if (metric != null) {
      builder.append(", ");
      builder.append("metric=").append(metric);
    }
    builder.append(", ");
    builder.append("labelValues=").append(labelValues);
    if (range != null) {
      builder.append(", ");
      builder.append("range=").append(range);
    }
    builder.append(", ");
    builder.append("withoutLabels=").append(withoutLabels);
    builder.append(", ");
    builder.append("byLabels=").append(byLabels);
    return builder.append("}").toString();
  }

  /**
   * Construct a new immutable {@code HttpQuery} instance.
   * @param host The value for the {@code host} attribute
   * @param function The value for the {@code function} attribute
   * @param aggregationOperator The value for the {@code aggregationOperator} attribute
   * @param aggregationOperatorParameters The value for the {@code aggregationOperatorParameters} attribute
   * @param metric The value for the {@code metric} attribute
   * @param labelValues The value for the {@code labelValues} attribute
   * @param range The value for the {@code range} attribute
   * @param withoutLabels The value for the {@code withoutLabels} attribute
   * @param byLabels The value for the {@code byLabels} attribute
   * @return An immutable HttpQuery instance
   */
  public static HttpQuery of(String host, Optional<String> function, Optional<String> aggregationOperator, List<String> aggregationOperatorParameters, String metric, List<HttpQuery.LabelValue> labelValues, Optional<String> range, List<String> withoutLabels, List<String> byLabels) {
    return of(host, function, aggregationOperator, (Iterable<String>) aggregationOperatorParameters, metric, (Iterable<? extends HttpQuery.LabelValue>) labelValues, range, (Iterable<String>) withoutLabels, (Iterable<String>) byLabels);
  }

  /**
   * Construct a new immutable {@code HttpQuery} instance.
   * @param host The value for the {@code host} attribute
   * @param function The value for the {@code function} attribute
   * @param aggregationOperator The value for the {@code aggregationOperator} attribute
   * @param aggregationOperatorParameters The value for the {@code aggregationOperatorParameters} attribute
   * @param metric The value for the {@code metric} attribute
   * @param labelValues The value for the {@code labelValues} attribute
   * @param range The value for the {@code range} attribute
   * @param withoutLabels The value for the {@code withoutLabels} attribute
   * @param byLabels The value for the {@code byLabels} attribute
   * @return An immutable HttpQuery instance
   */
  public static HttpQuery of(String host, Optional<String> function, Optional<String> aggregationOperator, Iterable<String> aggregationOperatorParameters, String metric, Iterable<? extends HttpQuery.LabelValue> labelValues, Optional<String> range, Iterable<String> withoutLabels, Iterable<String> byLabels) {
    return validate(new ImmutableHttpQuery(host, function, aggregationOperator, aggregationOperatorParameters, metric, labelValues, range, withoutLabels, byLabels));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableHttpQuery validate(ImmutableHttpQuery instance) {
    Set<ConstraintViolation<ImmutableHttpQuery>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link HttpQuery} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable HttpQuery instance
   */
  public static HttpQuery copyOf(HttpQuery instance) {
    if (instance instanceof ImmutableHttpQuery) {
      return (ImmutableHttpQuery) instance;
    }
    return new HttpQuery.Builder()
        .createFrom(instance)
        .build();
  }

  /**
   * Builds instances of type {@link HttpQuery HttpQuery}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  @Generated(from = "HttpQuery", generator = "Immutables")
  @NotThreadSafe
  public static class Builder {
    private @Nullable String host;
    private @Nullable String function;
    private @Nullable String aggregationOperator;
    private List<String> aggregationOperatorParameters = new ArrayList<String>();
    private @Nullable String metric;
    private List<HttpQuery.LabelValue> labelValues = new ArrayList<HttpQuery.LabelValue>();
    private @Nullable String range;
    private List<String> withoutLabels = new ArrayList<String>();
    private List<String> byLabels = new ArrayList<String>();

    /**
     * Creates a builder for {@link HttpQuery HttpQuery} instances.
     * <pre>
     * new HttpQuery.Builder()
     *    .host(String | null) // nullable {@link HttpQuery#getHost() host}
     *    .function(String) // optional {@link HttpQuery#getFunction() function}
     *    .aggregationOperator(String) // optional {@link HttpQuery#getAggregationOperator() aggregationOperator}
     *    .addAggregationOperatorParameter|addAllAggregationOperatorParameters(String) // {@link HttpQuery#getAggregationOperatorParameters() aggregationOperatorParameters} elements
     *    .metric(String | null) // nullable {@link HttpQuery#getMetric() metric}
     *    .addLabelValue|addAllLabelValues(io.syndesis.server.metrics.prometheus.HttpQuery.LabelValue) // {@link HttpQuery#getLabelValues() labelValues} elements
     *    .range(String) // optional {@link HttpQuery#getRange() range}
     *    .addWithoutLabel|addAllWithoutLabels(String) // {@link HttpQuery#getWithoutLabels() withoutLabels} elements
     *    .addByLabel|addAllByLabels(String) // {@link HttpQuery#getByLabels() byLabels} elements
     *    .build();
     * </pre>
     */
    public Builder() {
      if (!(this instanceof HttpQuery.Builder)) {
        throw new UnsupportedOperationException("Use: new HttpQuery.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code HttpQuery} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder createFrom(HttpQuery instance) {
      Objects.requireNonNull(instance, "instance");
      String hostValue = instance.getHost();
      if (hostValue != null) {
        host(hostValue);
      }
      Optional<String> functionOptional = instance.getFunction();
      if (functionOptional.isPresent()) {
        function(functionOptional);
      }
      Optional<String> aggregationOperatorOptional = instance.getAggregationOperator();
      if (aggregationOperatorOptional.isPresent()) {
        aggregationOperator(aggregationOperatorOptional);
      }
      addAllAggregationOperatorParameters(instance.getAggregationOperatorParameters());
      String metricValue = instance.getMetric();
      if (metricValue != null) {
        metric(metricValue);
      }
      addAllLabelValues(instance.getLabelValues());
      Optional<String> rangeOptional = instance.getRange();
      if (rangeOptional.isPresent()) {
        range(rangeOptional);
      }
      addAllWithoutLabels(instance.getWithoutLabels());
      addAllByLabels(instance.getByLabels());
      return (HttpQuery.Builder) this;
    }

    /**
     * Initializes the value for the {@link HttpQuery#getHost() host} attribute.
     * @param host The value for host (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder host(String host) {
      this.host = host;
      return (HttpQuery.Builder) this;
    }

    /**
     * Initializes the optional value {@link HttpQuery#getFunction() function} to function.
     * @param function The value for function
     * @return {@code this} builder for chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder function(String function) {
      this.function = Objects.requireNonNull(function, "function");
      return (HttpQuery.Builder) this;
    }

    /**
     * Initializes the optional value {@link HttpQuery#getFunction() function} to function.
     * @param function The value for function
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder function(Optional<String> function) {
      this.function = function.orElse(null);
      return (HttpQuery.Builder) this;
    }

    /**
     * Initializes the optional value {@link HttpQuery#getAggregationOperator() aggregationOperator} to aggregationOperator.
     * @param aggregationOperator The value for aggregationOperator
     * @return {@code this} builder for chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder aggregationOperator(String aggregationOperator) {
      this.aggregationOperator = Objects.requireNonNull(aggregationOperator, "aggregationOperator");
      return (HttpQuery.Builder) this;
    }

    /**
     * Initializes the optional value {@link HttpQuery#getAggregationOperator() aggregationOperator} to aggregationOperator.
     * @param aggregationOperator The value for aggregationOperator
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder aggregationOperator(Optional<String> aggregationOperator) {
      this.aggregationOperator = aggregationOperator.orElse(null);
      return (HttpQuery.Builder) this;
    }

    /**
     * Adds one element to {@link HttpQuery#getAggregationOperatorParameters() aggregationOperatorParameters} list.
     * @param element A aggregationOperatorParameters element
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addAggregationOperatorParameter(String element) {
      this.aggregationOperatorParameters.add(Objects.requireNonNull(element, "aggregationOperatorParameters element"));
      return (HttpQuery.Builder) this;
    }

    /**
     * Adds elements to {@link HttpQuery#getAggregationOperatorParameters() aggregationOperatorParameters} list.
     * @param elements An array of aggregationOperatorParameters elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addAggregationOperatorParameters(String... elements) {
      for (String element : elements) {
        this.aggregationOperatorParameters.add(Objects.requireNonNull(element, "aggregationOperatorParameters element"));
      }
      return (HttpQuery.Builder) this;
    }


    /**
     * Sets or replaces all elements for {@link HttpQuery#getAggregationOperatorParameters() aggregationOperatorParameters} list.
     * @param elements An iterable of aggregationOperatorParameters elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder aggregationOperatorParameters(Iterable<String> elements) {
      this.aggregationOperatorParameters.clear();
      return addAllAggregationOperatorParameters(elements);
    }

    /**
     * Adds elements to {@link HttpQuery#getAggregationOperatorParameters() aggregationOperatorParameters} list.
     * @param elements An iterable of aggregationOperatorParameters elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addAllAggregationOperatorParameters(Iterable<String> elements) {
      for (String element : elements) {
        this.aggregationOperatorParameters.add(Objects.requireNonNull(element, "aggregationOperatorParameters element"));
      }
      return (HttpQuery.Builder) this;
    }

    /**
     * Initializes the value for the {@link HttpQuery#getMetric() metric} attribute.
     * @param metric The value for metric (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder metric(String metric) {
      this.metric = metric;
      return (HttpQuery.Builder) this;
    }

    /**
     * Adds one element to {@link HttpQuery#getLabelValues() labelValues} list.
     * @param element A labelValues element
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addLabelValue(HttpQuery.LabelValue element) {
      this.labelValues.add(Objects.requireNonNull(element, "labelValues element"));
      return (HttpQuery.Builder) this;
    }

    /**
     * Adds elements to {@link HttpQuery#getLabelValues() labelValues} list.
     * @param elements An array of labelValues elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addLabelValues(HttpQuery.LabelValue... elements) {
      for (HttpQuery.LabelValue element : elements) {
        this.labelValues.add(Objects.requireNonNull(element, "labelValues element"));
      }
      return (HttpQuery.Builder) this;
    }


    /**
     * Sets or replaces all elements for {@link HttpQuery#getLabelValues() labelValues} list.
     * @param elements An iterable of labelValues elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder labelValues(Iterable<? extends HttpQuery.LabelValue> elements) {
      this.labelValues.clear();
      return addAllLabelValues(elements);
    }

    /**
     * Adds elements to {@link HttpQuery#getLabelValues() labelValues} list.
     * @param elements An iterable of labelValues elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addAllLabelValues(Iterable<? extends HttpQuery.LabelValue> elements) {
      for (HttpQuery.LabelValue element : elements) {
        this.labelValues.add(Objects.requireNonNull(element, "labelValues element"));
      }
      return (HttpQuery.Builder) this;
    }

    /**
     * Initializes the optional value {@link HttpQuery#getRange() range} to range.
     * @param range The value for range
     * @return {@code this} builder for chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder range(String range) {
      this.range = Objects.requireNonNull(range, "range");
      return (HttpQuery.Builder) this;
    }

    /**
     * Initializes the optional value {@link HttpQuery#getRange() range} to range.
     * @param range The value for range
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder range(Optional<String> range) {
      this.range = range.orElse(null);
      return (HttpQuery.Builder) this;
    }

    /**
     * Adds one element to {@link HttpQuery#getWithoutLabels() withoutLabels} list.
     * @param element A withoutLabels element
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addWithoutLabel(String element) {
      this.withoutLabels.add(Objects.requireNonNull(element, "withoutLabels element"));
      return (HttpQuery.Builder) this;
    }

    /**
     * Adds elements to {@link HttpQuery#getWithoutLabels() withoutLabels} list.
     * @param elements An array of withoutLabels elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addWithoutLabels(String... elements) {
      for (String element : elements) {
        this.withoutLabels.add(Objects.requireNonNull(element, "withoutLabels element"));
      }
      return (HttpQuery.Builder) this;
    }


    /**
     * Sets or replaces all elements for {@link HttpQuery#getWithoutLabels() withoutLabels} list.
     * @param elements An iterable of withoutLabels elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder withoutLabels(Iterable<String> elements) {
      this.withoutLabels.clear();
      return addAllWithoutLabels(elements);
    }

    /**
     * Adds elements to {@link HttpQuery#getWithoutLabels() withoutLabels} list.
     * @param elements An iterable of withoutLabels elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addAllWithoutLabels(Iterable<String> elements) {
      for (String element : elements) {
        this.withoutLabels.add(Objects.requireNonNull(element, "withoutLabels element"));
      }
      return (HttpQuery.Builder) this;
    }

    /**
     * Adds one element to {@link HttpQuery#getByLabels() byLabels} list.
     * @param element A byLabels element
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addByLabel(String element) {
      this.byLabels.add(Objects.requireNonNull(element, "byLabels element"));
      return (HttpQuery.Builder) this;
    }

    /**
     * Adds elements to {@link HttpQuery#getByLabels() byLabels} list.
     * @param elements An array of byLabels elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addByLabels(String... elements) {
      for (String element : elements) {
        this.byLabels.add(Objects.requireNonNull(element, "byLabels element"));
      }
      return (HttpQuery.Builder) this;
    }


    /**
     * Sets or replaces all elements for {@link HttpQuery#getByLabels() byLabels} list.
     * @param elements An iterable of byLabels elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder byLabels(Iterable<String> elements) {
      this.byLabels.clear();
      return addAllByLabels(elements);
    }

    /**
     * Adds elements to {@link HttpQuery#getByLabels() byLabels} list.
     * @param elements An iterable of byLabels elements
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.Builder addAllByLabels(Iterable<String> elements) {
      for (String element : elements) {
        this.byLabels.add(Objects.requireNonNull(element, "byLabels element"));
      }
      return (HttpQuery.Builder) this;
    }

    /**
     * Builds a new {@link HttpQuery HttpQuery}.
     * @return An immutable instance of HttpQuery
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public HttpQuery build() {
      return ImmutableHttpQuery.validate(new ImmutableHttpQuery(
          null,
          host,
          function,
          aggregationOperator,
          createUnmodifiableList(true, aggregationOperatorParameters),
          metric,
          createUnmodifiableList(true, labelValues),
          range,
          createUnmodifiableList(true, withoutLabels),
          createUnmodifiableList(true, byLabels)));
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }
}
