/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.server.metrics.prometheus;

import com.google.errorprone.annotations.CanIgnoreReturnValue;
import com.google.errorprone.annotations.Var;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import java.util.Objects;
import java.util.Set;
import javax.annotation.CheckReturnValue;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import javax.annotation.concurrent.Immutable;
import javax.annotation.concurrent.NotThreadSafe;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;
import org.immutables.value.Generated;

/**
 * Immutable implementation of {@link HttpQuery.LabelValue}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new HttpQuery.LabelValue.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableLabelValue.of()}.
 */
@Generated(from = "HttpQuery.LabelValue", generator = "Immutables")
@SuppressWarnings({"all"})
@SuppressFBWarnings
@ParametersAreNonnullByDefault
@javax.annotation.Generated("org.immutables.processor.ProxyProcessor")
@Immutable
@CheckReturnValue
final class ImmutableLabelValue implements HttpQuery.LabelValue {
  private final String label;
  private final String value;

  private ImmutableLabelValue(String label, String value) {
    this.label = label;
    this.value = value;
  }

  /**
   * @return The value of the {@code label} attribute
   */
  @Override
  public String getLabel() {
    return label;
  }

  /**
   * @return The value of the {@code value} attribute
   */
  @Override
  public String getValue() {
    return value;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link HttpQuery.LabelValue#getLabel() label} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for label (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableLabelValue withLabel(String value) {
    if (Objects.equals(this.label, value)) return this;
    return validate(new ImmutableLabelValue(value, this.value));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link HttpQuery.LabelValue#getValue() value} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for value (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableLabelValue withValue(String value) {
    if (Objects.equals(this.value, value)) return this;
    return validate(new ImmutableLabelValue(this.label, value));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableLabelValue} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(@Nullable Object another) {
    if (this == another) return true;
    return another instanceof ImmutableLabelValue
        && equalTo((ImmutableLabelValue) another);
  }

  private boolean equalTo(ImmutableLabelValue another) {
    return Objects.equals(label, another.label)
        && Objects.equals(value, another.value);
  }

  /**
   * Computes a hash code from attributes: {@code label}, {@code value}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    @Var int h = 5381;
    h += (h << 5) + Objects.hashCode(label);
    h += (h << 5) + Objects.hashCode(value);
    return h;
  }


  /**
   * Prints the immutable value {@code LabelValue} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "LabelValue{"
        + "label=" + label
        + ", value=" + value
        + "}";
  }

  /**
   * Construct a new immutable {@code LabelValue} instance.
   * @param label The value for the {@code label} attribute
   * @param value The value for the {@code value} attribute
   * @return An immutable LabelValue instance
   */
  public static HttpQuery.LabelValue of(String label, String value) {
    return validate(new ImmutableLabelValue(label, value));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableLabelValue validate(ImmutableLabelValue instance) {
    Set<ConstraintViolation<ImmutableLabelValue>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link HttpQuery.LabelValue} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable LabelValue instance
   */
  public static HttpQuery.LabelValue copyOf(HttpQuery.LabelValue instance) {
    if (instance instanceof ImmutableLabelValue) {
      return (ImmutableLabelValue) instance;
    }
    return new HttpQuery.LabelValue.Builder()
        .createFrom(instance)
        .build();
  }

  /**
   * Builds instances of type {@link HttpQuery.LabelValue LabelValue}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  @Generated(from = "HttpQuery.LabelValue", generator = "Immutables")
  @NotThreadSafe
  public static class Builder {
    private @Nullable String label;
    private @Nullable String value;

    /**
     * Creates a builder for {@link HttpQuery.LabelValue LabelValue} instances.
     * <pre>
     * new HttpQuery.LabelValue.Builder()
     *    .label(String | null) // nullable {@link HttpQuery.LabelValue#getLabel() label}
     *    .value(String | null) // nullable {@link HttpQuery.LabelValue#getValue() value}
     *    .build();
     * </pre>
     */
    public Builder() {
      if (!(this instanceof HttpQuery.LabelValue.Builder)) {
        throw new UnsupportedOperationException("Use: new HttpQuery.LabelValue.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code LabelValue} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.LabelValue.Builder createFrom(HttpQuery.LabelValue instance) {
      Objects.requireNonNull(instance, "instance");
      String labelValue = instance.getLabel();
      if (labelValue != null) {
        label(labelValue);
      }
      String valueValue = instance.getValue();
      if (valueValue != null) {
        value(valueValue);
      }
      return (HttpQuery.LabelValue.Builder) this;
    }

    /**
     * Initializes the value for the {@link HttpQuery.LabelValue#getLabel() label} attribute.
     * @param label The value for label (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.LabelValue.Builder label(String label) {
      this.label = label;
      return (HttpQuery.LabelValue.Builder) this;
    }

    /**
     * Initializes the value for the {@link HttpQuery.LabelValue#getValue() value} attribute.
     * @param value The value for value (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final HttpQuery.LabelValue.Builder value(String value) {
      this.value = value;
      return (HttpQuery.LabelValue.Builder) this;
    }

    /**
     * Builds a new {@link HttpQuery.LabelValue LabelValue}.
     * @return An immutable instance of LabelValue
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public HttpQuery.LabelValue build() {
      return ImmutableLabelValue.validate(new ImmutableLabelValue(label, value));
    }
  }
}
