/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.server.metrics.prometheus;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.errorprone.annotations.CanIgnoreReturnValue;
import com.google.errorprone.annotations.Var;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import java.io.ObjectStreamException;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import javax.annotation.CheckReturnValue;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import javax.annotation.concurrent.Immutable;
import javax.annotation.concurrent.NotThreadSafe;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;
import org.immutables.value.Generated;

/**
 * Immutable implementation of {@link QueryResult}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new QueryResult.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableQueryResult.of()}.
 */
@Generated(from = "QueryResult", generator = "Immutables")
@SuppressWarnings({"all"})
@SuppressFBWarnings
@ParametersAreNonnullByDefault
@javax.annotation.Generated("org.immutables.processor.ProxyProcessor")
@Immutable
@CheckReturnValue
final class ImmutableQueryResult implements QueryResult {
  private final String status;
  private final @Nullable String errorType;
  private final @Nullable String error;
  private final @Nullable QueryResult.Data data;

  @SuppressWarnings("unchecked") // safe covariant cast
  private ImmutableQueryResult(
      String status,
      Optional<String> errorType,
      Optional<String> error,
      Optional<? extends QueryResult.Data> data) {
    this.status = status;
    this.errorType = errorType.orElse(null);
    this.error = error.orElse(null);
    this.data = data.orElse(null);
  }

  private ImmutableQueryResult(
      ImmutableQueryResult original,
      String status,
      @Nullable String errorType,
      @Nullable String error,
      @Nullable QueryResult.Data data) {
    this.status = status;
    this.errorType = errorType;
    this.error = error;
    this.data = data;
  }

  /**
   * @return The value of the {@code status} attribute
   */
  @JsonProperty("status")
  @Override
  public String getStatus() {
    return status;
  }

  /**
   * @return The value of the {@code errorType} attribute
   */
  @JsonProperty("errorType")
  @Override
  public Optional<String> getErrorType() {
    return Optional.ofNullable(errorType);
  }

  /**
   * @return The value of the {@code error} attribute
   */
  @JsonProperty("error")
  @Override
  public Optional<String> getError() {
    return Optional.ofNullable(error);
  }

  /**
   * @return The value of the {@code data} attribute
   */
  @JsonProperty("data")
  @Override
  public Optional<QueryResult.Data> getData() {
    return Optional.ofNullable(data);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link QueryResult#getStatus() status} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for status (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableQueryResult withStatus(String value) {
    if (Objects.equals(this.status, value)) return this;
    return validate(new ImmutableQueryResult(this, value, this.errorType, this.error, this.data));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link QueryResult#getErrorType() errorType} attribute.
   * @param value The value for errorType
   * @return A modified copy of {@code this} object
   */
  public final ImmutableQueryResult withErrorType(String value) {
    @Nullable String newValue = Objects.requireNonNull(value, "errorType");
    if (Objects.equals(this.errorType, newValue)) return this;
    return validate(new ImmutableQueryResult(this, this.status, newValue, this.error, this.data));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link QueryResult#getErrorType() errorType} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for errorType
   * @return A modified copy of {@code this} object
   */
  public final ImmutableQueryResult withErrorType(Optional<String> optional) {
    @Nullable String value = optional.orElse(null);
    if (Objects.equals(this.errorType, value)) return this;
    return validate(new ImmutableQueryResult(this, this.status, value, this.error, this.data));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link QueryResult#getError() error} attribute.
   * @param value The value for error
   * @return A modified copy of {@code this} object
   */
  public final ImmutableQueryResult withError(String value) {
    @Nullable String newValue = Objects.requireNonNull(value, "error");
    if (Objects.equals(this.error, newValue)) return this;
    return validate(new ImmutableQueryResult(this, this.status, this.errorType, newValue, this.data));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link QueryResult#getError() error} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for error
   * @return A modified copy of {@code this} object
   */
  public final ImmutableQueryResult withError(Optional<String> optional) {
    @Nullable String value = optional.orElse(null);
    if (Objects.equals(this.error, value)) return this;
    return validate(new ImmutableQueryResult(this, this.status, this.errorType, value, this.data));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link QueryResult#getData() data} attribute.
   * @param value The value for data
   * @return A modified copy of {@code this} object
   */
  public final ImmutableQueryResult withData(QueryResult.Data value) {
    @Nullable QueryResult.Data newValue = Objects.requireNonNull(value, "data");
    if (this.data == newValue) return this;
    return validate(new ImmutableQueryResult(this, this.status, this.errorType, this.error, newValue));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link QueryResult#getData() data} attribute.
   * A shallow reference equality check is used on unboxed optional value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for data
   * @return A modified copy of {@code this} object
   */
  @SuppressWarnings("unchecked") // safe covariant cast
  public final ImmutableQueryResult withData(Optional<? extends QueryResult.Data> optional) {
    @Nullable QueryResult.Data value = optional.orElse(null);
    if (this.data == value) return this;
    return validate(new ImmutableQueryResult(this, this.status, this.errorType, this.error, value));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableQueryResult} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(@Nullable Object another) {
    if (this == another) return true;
    return another instanceof ImmutableQueryResult
        && equalTo((ImmutableQueryResult) another);
  }

  private boolean equalTo(ImmutableQueryResult another) {
    return Objects.equals(status, another.status)
        && Objects.equals(errorType, another.errorType)
        && Objects.equals(error, another.error)
        && Objects.equals(data, another.data);
  }

  /**
   * Computes a hash code from attributes: {@code status}, {@code errorType}, {@code error}, {@code data}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    @Var int h = 5381;
    h += (h << 5) + Objects.hashCode(status);
    h += (h << 5) + Objects.hashCode(errorType);
    h += (h << 5) + Objects.hashCode(error);
    h += (h << 5) + Objects.hashCode(data);
    return h;
  }


  /**
   * Prints the immutable value {@code QueryResult} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder("QueryResult{");
    if (status != null) {
      builder.append("status=").append(status);
    }
    if (errorType != null) {
      if (builder.length() > 12) builder.append(", ");
      builder.append("errorType=").append(errorType);
    }
    if (error != null) {
      if (builder.length() > 12) builder.append(", ");
      builder.append("error=").append(error);
    }
    if (data != null) {
      if (builder.length() > 12) builder.append(", ");
      builder.append("data=").append(data);
    }
    return builder.append("}").toString();
  }

  /**
   * Construct a new immutable {@code QueryResult} instance.
   * @param status The value for the {@code status} attribute
   * @param errorType The value for the {@code errorType} attribute
   * @param error The value for the {@code error} attribute
   * @param data The value for the {@code data} attribute
   * @return An immutable QueryResult instance
   */
  public static QueryResult of(String status, Optional<String> errorType, Optional<String> error, Optional<? extends QueryResult.Data> data) {
    return validate(new ImmutableQueryResult(status, errorType, error, data));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableQueryResult validate(ImmutableQueryResult instance) {
    Set<ConstraintViolation<ImmutableQueryResult>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link QueryResult} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable QueryResult instance
   */
  public static QueryResult copyOf(QueryResult instance) {
    if (instance instanceof ImmutableQueryResult) {
      return (ImmutableQueryResult) instance;
    }
    return new QueryResult.Builder()
        .createFrom(instance)
        .build();
  }

  private Object readResolve() throws ObjectStreamException {
    return validate(this);
  }

  /**
   * Builds instances of type {@link QueryResult QueryResult}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  @Generated(from = "QueryResult", generator = "Immutables")
  @NotThreadSafe
  public static class Builder {
    private @Nullable String status;
    private @Nullable String errorType;
    private @Nullable String error;
    private @Nullable QueryResult.Data data;

    /**
     * Creates a builder for {@link QueryResult QueryResult} instances.
     * <pre>
     * new QueryResult.Builder()
     *    .status(String | null) // nullable {@link QueryResult#getStatus() status}
     *    .errorType(String) // optional {@link QueryResult#getErrorType() errorType}
     *    .error(String) // optional {@link QueryResult#getError() error}
     *    .data(io.syndesis.server.metrics.prometheus.QueryResult.Data) // optional {@link QueryResult#getData() data}
     *    .build();
     * </pre>
     */
    public Builder() {
      if (!(this instanceof QueryResult.Builder)) {
        throw new UnsupportedOperationException("Use: new QueryResult.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code QueryResult} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Builder createFrom(QueryResult instance) {
      Objects.requireNonNull(instance, "instance");
      String statusValue = instance.getStatus();
      if (statusValue != null) {
        status(statusValue);
      }
      Optional<String> errorTypeOptional = instance.getErrorType();
      if (errorTypeOptional.isPresent()) {
        errorType(errorTypeOptional);
      }
      Optional<String> errorOptional = instance.getError();
      if (errorOptional.isPresent()) {
        error(errorOptional);
      }
      Optional<QueryResult.Data> dataOptional = instance.getData();
      if (dataOptional.isPresent()) {
        data(dataOptional);
      }
      return (QueryResult.Builder) this;
    }

    /**
     * Initializes the value for the {@link QueryResult#getStatus() status} attribute.
     * @param status The value for status (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    @JsonProperty("status")
    public final QueryResult.Builder status(String status) {
      this.status = status;
      return (QueryResult.Builder) this;
    }

    /**
     * Initializes the optional value {@link QueryResult#getErrorType() errorType} to errorType.
     * @param errorType The value for errorType
     * @return {@code this} builder for chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Builder errorType(String errorType) {
      this.errorType = Objects.requireNonNull(errorType, "errorType");
      return (QueryResult.Builder) this;
    }

    /**
     * Initializes the optional value {@link QueryResult#getErrorType() errorType} to errorType.
     * @param errorType The value for errorType
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    @JsonProperty("errorType")
    public final QueryResult.Builder errorType(Optional<String> errorType) {
      this.errorType = errorType.orElse(null);
      return (QueryResult.Builder) this;
    }

    /**
     * Initializes the optional value {@link QueryResult#getError() error} to error.
     * @param error The value for error
     * @return {@code this} builder for chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Builder error(String error) {
      this.error = Objects.requireNonNull(error, "error");
      return (QueryResult.Builder) this;
    }

    /**
     * Initializes the optional value {@link QueryResult#getError() error} to error.
     * @param error The value for error
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    @JsonProperty("error")
    public final QueryResult.Builder error(Optional<String> error) {
      this.error = error.orElse(null);
      return (QueryResult.Builder) this;
    }

    /**
     * Initializes the optional value {@link QueryResult#getData() data} to data.
     * @param data The value for data
     * @return {@code this} builder for chained invocation
     */
    @CanIgnoreReturnValue 
    public final QueryResult.Builder data(QueryResult.Data data) {
      this.data = Objects.requireNonNull(data, "data");
      return (QueryResult.Builder) this;
    }

    /**
     * Initializes the optional value {@link QueryResult#getData() data} to data.
     * @param data The value for data
     * @return {@code this} builder for use in a chained invocation
     */
    @CanIgnoreReturnValue 
    @JsonProperty("data")
    public final QueryResult.Builder data(Optional<? extends QueryResult.Data> data) {
      this.data = data.orElse(null);
      return (QueryResult.Builder) this;
    }

    /**
     * Builds a new {@link QueryResult QueryResult}.
     * @return An immutable instance of QueryResult
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public QueryResult build() {
      return ImmutableQueryResult.validate(new ImmutableQueryResult(null, status, errorType, error, data));
    }
  }
}
